'''
Window based evaluation metricks package.  Provides window based segmentation
evaluation metrics including:

* Pk 
* WindowDiff and
* WinPR

.. warning:: These are provided for comparison, but are not recommended for \
    segmentation evaluation.  Instead, use  the segmentation similarity
    metric [FournierInkpen2012]_ implemented in
    :func:`segeval.similarity.SegmentationSimilarity.similarity` and the
    associated inter-coder agreement coefficients in
    :mod:`segeval.agreement`.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal
from ..Math import mean, std, var
from .. import convert_positions_to_masses


def load_tests(loader, tests, pattern):
    '''
    A ``load_tests()`` function utilizing the default loader
    :func:`segeval.Utils.default_load_tests`.
    
    .. seealso:: The `load_tests protocol <http://docs.python.org/library/\
    unittest.html#load-tests-protocol>`_.
    '''
    #pylint: disable=W0613
    from ..Utils import default_load_tests
    return default_load_tests(__file__, loader, tests)


def compute_window_size_from_masses(coder_masses, fnc_round=round):
    '''
    Compute a window size from a dict of segment masses.
    
    :param coder_masses: A dict of segment masses.
    :type coder_masses: dict
    '''
    masses = list()
    # Define fnc
    def __list_coder_masses__(inner_coder_masses):
        '''
        Recursively collect all masses.
        
        :param inner_coder_masses: Either a dict of dicts, or dict of a list of
            masses.
        :type inner_coder_masses: dict or list
        '''
        if isinstance(inner_coder_masses, list):
            masses.extend(inner_coder_masses)
        elif isinstance(inner_coder_masses, dict):
            for cur_inner_coder_masses in inner_coder_masses.values():
                __list_coder_masses__(cur_inner_coder_masses)
    # Recurse and list all masses
    __list_coder_masses__(coder_masses)
    # Convert to floats
    masses = [Decimal(mass) for mass in masses]
    # Calculate
    avg = mean(masses) / 2.0
    window_size = int(fnc_round(avg))
    return window_size if window_size > 1 else 2
    

def compute_window_size(reference_segments, fnc_round=round):
    '''
    Compute the window size to use with window_diff from half of the average
    segment length in the reference segmentation.
    
    :param ref_segments: An ordered sequence of which section each unit belongs
        to, e.g., ``[1,1,1,1,1,2,2,2,3,3,3,3,3]`` for 13 units (e.g., sentences,
        paragraphs)
    :param fnc_round: rounding function to use (default: :func:`round`).  
        WinDiff is very sensitive to window size, so to reproduce results from
        another implementation another rounding function may need to be
        specified (e.g. :func:`math.ceil`, :func:`math.floor`, etc.).
    
    :returns: Integer window size to use with win_diff.
    :rtype: int
    '''
    masses = convert_positions_to_masses(reference_segments)
    masses = [Decimal(mass) for mass in masses]
    avg = mean(masses) / Decimal(2)
    window_size = int(fnc_round(avg))
    return window_size if window_size > 1 else 2


def parser_one_minus_support(parser):
    '''
    Add support for the "one minus" parameter to convert penalty metrics into
    reward metrics.
    
    :param parser: Argument parser
    :type parser: argparse.ArgumentParser
    '''
    parser.add_argument('-om', '--oneminus',
                        default=False,
                        action='store_true',
                        help='Calculates 1-metric to make this metric no \
                                longer penalty-based, meaning that 1.0 \
                                represents the best performance and 0.0 the \
                                worst.')
    
    


'''
Display and output related utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================


def values_to_str(mean, std, var, stderr):
    '''
    Create a text representation of a metric's mean.
    '''
    return ('\tmean\t= %(mean)s\n'+\
            '\tstd\t= %(std)s\n'+\
            '\tvar\t= %(var)s\n'+\
            '\tstderr\t= %(stderr)s') % \
            {'mean'   : mean,
             'std'    : std,
             'var'    : var,
             'stderr' : stderr}


def render_mean_values(name, mean, std, var, stderr):
    '''
    Render text representing means of a metric.
    '''
    return render_value(name, values_to_str(mean, std, var, stderr),
                        operator='\n')


def render_value(name, value, operator='='):
    '''
    Render text representing the value of a metric.
    '''
    return '%(name)s %(operator)s %(value)s' % {'name' : name,
                                                'value' : value,
                                                'operator' : operator}





    


'''
Tests linear boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from .SingleBoundaryDistance import linear_edit_distance


class TestLinearBoundaryDistance(unittest.TestCase):
    '''
    Tests linear boundary edit distance.
    '''
    # pylint: disable=R0904,C0324,C0103
    
    SKIP = False

    def test_nested_transp_case_linear(self):
        '''
        Test the linear transposition case.
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [1,1,7]
        mass_b = [5,1,3]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     7)
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     5)
        
        self.assertEqual((d,len(t),len(s)), (3,1,2))


    def test_suboptimal_transp(self):
        '''
        Test a decision to use substitutions instead of a transposition.
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [1,3,2]
        mass_b = [2,3,1]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     4)
        self.assertEqual((d,len(t),len(s)), (2,2,0))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     2)
        self.assertEqual((d,len(t),len(s)), (2,2,0))


    def test_n2_transp(self):
        '''
        Test transpositions with n=2
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [5,5]
        mass_b = [6,4]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     1)
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     2)
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     3)
        self.assertEqual((d,len(t),len(s)), (1,1,0))


    def test_n3_transp(self):
        '''
        Test transpositions with n=3
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [5,5]
        mass_b = [7,3]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     1)
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     2)
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     3)
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     4)
        self.assertEqual((d,len(t),len(s)), (1,1,0))

    
    def test_complex_case(self):
        '''
        Test a combination of substitutions and transpositions
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [1,2,2,4]
        mass_b = [2,2,5]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     4)
        self.assertEqual((d,len(t),len(s)), (3,2,1))
        
    def test_distance_1(self):
        '''
        Test a distance of one.
        '''
        if TestLinearBoundaryDistance.SKIP:
            return
        
        mass_a = [1,4,4]
        mass_b = [1,2,6]
        
        d,t,s = linear_edit_distance(mass_a,
                                     mass_b,
                                     5)
                                     
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        
        
        
        
        
        


'''
Provides a segmentation version of the F-Measure metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from .Percentage import find_boundary_position_freqs
from . import fmeasure, precision as ml_precision, recall as ml_recall
from .. import compute_pairwise, compute_pairwise_values, create_tsv_rows
from ..data import load_file
from ..data.TSV import write_tsv
from ..data.Display import render_mean_values


DEFAULT_BETA = 1.0


def f_b_measure(hypothesis_masses, reference_masses, beta=DEFAULT_BETA,
                return_parts=False):
    '''
    Calculates the F-Measure between a hypothesis and reference segmentation,
    where F-Measure is calculated as:
    
    .. math::
        \\text{F}_{\\beta}\\text{-measure} = \\frac{(1 + \\beta^2) \\cdot TP}\
        {(1 + \\beta^2) \\cdot TP + \\beta^2 \\cdot FN + FP}
    
    Counts of true positives (:math:`TP`), false positives (:math:`FP`), and
    false negatives (:math:`FN`) are calculated as:
    
    .. math::
        TP = \\sum^{|hyp|}_{i=1}{\\text{tp}(hyp_i, ref_i)}, \quad
        FP = \\sum^{|hyp|}_{i=1}{\\text{fp}(hyp_i, ref_i)}, \quad
        FN = \\sum^{|hyp|}_{i=1}{\\text{fn}(hyp_i, ref_i)}
        
    .. math::
        \\text{tp}(hyp_i, ref_i) = 
        \\begin{cases}
            1    & \\text{if both } hyp_i \\text{ and } ref_i \\text{ are boundaries}  \\\\
            0    & \\text{else}
        \\end{cases}
        
    .. math::
        \\text{fp}(hyp_i, ref_i) = 
        \\begin{cases}
            1    & \\text{if } hyp_i \\text{ is a boundary and } ref_i \\text{ is not}  \\\\
            0    & \\text{else}
        \\end{cases}
        
    .. math::
        \\text{fn}(hyp_i, ref_i) = 
        \\begin{cases}
            1    & \\text{if is not a boundary } hyp_i \\text{ and } ref_i \\text{ is}  \\\\
            0    & \\text{else}
        \\end{cases}
        
    .. math::
        \\text{tn}(hyp_i, ref_i) = 
        \\begin{cases}
            1    & \\text{if both } hyp_i \\text{ and } ref_i \\text{ are not boundaries}  \\\\
            0    & \\text{else}
        \\end{cases}
    
    Each matching boundary position is considered a TP, whereas a missing
    boundary in the hypothesis is considered a FN, and an extra boundary in the
    hypothesis that is not found in the reference is considered a FP.  TNs
    do not occur.
    
    :param hypothesis_masses: Hypothesis segmentation masses
    :param reference_masses:  Reference segmentation masses
    :param beta:              Scales how precision and recall are averaged
    :type hypothesis_masses: list
    :type reference_masses:  list
    :type beta:              :func:`float` or :class:`decimal.Decimal`
    
    :returns: F-measure or the values of the confusion matrix
    :rtype: :class:`decimal.Decimal` or :func:`int`, :func:`int`, :func:`int`, \
        :func:`int`
    
    .. seealso:: :func:`segeval.ml.fmeasure`
    '''
    # pylint: disable=C0103
    positions_hyp = find_boundary_position_freqs([hypothesis_masses])
    positions_ref = find_boundary_position_freqs([reference_masses])
    tp = 0
    fp = 0
    fn = 0
    for pos in positions_ref.keys():
        if pos in positions_hyp:
            tp += 1
        else:
            fn += 1
    for pos in positions_hyp.keys():
        if pos not in positions_ref:
            fp += 1
    
    tn = sum(reference_masses) - 1 - len(positions_ref) - fn - fp
    
    if not return_parts:
        return fmeasure(tp, fp, fn, beta)
    else:
        return tp, fp, fn, tn
        


def precision(hypothesis_masses, reference_masses):
    '''
    Calculates the precision between a hypothesis and reference segmentation,
    where precision is calculated as:
    
    .. math::
        \\text{Precision} = \\frac{TP}{TP + FP}
    
    Counts of true positives (:math:`TP`) and false negatives (:math:`FP`) are
    calculated as in :func:`f_b_measure`.
    
    :param hypothesis_masses: Hypothesis segmentation masses.
    :param reference_masses: Reference segmentation masses.
    :type hypothesis_masses: list
    :type reference_masses: list
    
    :returns: Precision.
    :rtype: :class:`decimal.Decimal`
    
    .. seealso:: :func:`segeval.ml.precision`
    '''
    # pylint: disable=C0103
    positions_hyp = find_boundary_position_freqs([hypothesis_masses])
    positions_ref = find_boundary_position_freqs([reference_masses])
    tp = 0
    fp = 0
    for pos in positions_ref.keys():
        if pos in positions_hyp:
            tp += 1
    for pos in positions_hyp.keys():
        if pos not in positions_ref:
            fp += 1
    return ml_precision(tp, fp)


def recall(hypothesis_masses, reference_masses):
    '''
    Calculates the recall between a hypothesis and reference segmentation,
    where recall is calculated as:
    
    .. math::
        \\text{Recall} = \\frac{TP}{TP + FN}
    
    Counts of true positives (:math:`TP`) and false negatives (:math:`FN`) are
    calculated as in :func:`f_b_measure`.
    
    :param hypothesis_masses: Hypothesis segmentation masses.
    :param reference_masses: Reference segmentation masses.
    :type hypothesis_masses: list
    :type reference_masses: list
    
    :returns: Precision.
    :rtype: :class:`decimal.Decimal`
    
    .. seealso:: :func:`segeval.ml.precision`
    '''
    # pylint: disable=C0103
    positions_hyp = find_boundary_position_freqs([hypothesis_masses])
    positions_ref = find_boundary_position_freqs([reference_masses])
    tp = 0
    fn = 0
    for pos in positions_ref.keys():
        if pos in positions_hyp:
            tp += 1
        else:
            fn += 1
    return ml_recall(tp, fn)


def pairwise_f_b_measure(dataset_masses, beta=DEFAULT_BETA, permuted=True):
    '''
    Calculate mean pairwise segmentation F-Measure.
    
    .. seealso:: :func:`f_b_measure`
    .. seealso:: :func:`segeval.compute_pairwise`
    
    :param dataset_masses: Segmentation mass dataset (including multiple \
                           codings).
    :type dataset_masses: dict
        
    :returns: Mean, standard deviation, variance, and standard error of a \
        segmentation metric.
    :rtype: :class:`decimal.Decimal`, :class:`decimal.Decimal`, \
        :class:`decimal.Decimal`, :class:`decimal.Decimal`
    '''
    def wrapper(hypothesis_masses, reference_masses):
        '''
        Wrapper to provide parameters.
        '''
        return f_b_measure(hypothesis_masses, reference_masses, beta)
    return compute_pairwise(dataset_masses, wrapper, permuted=permuted)


def pairwise_ml_measure(dataset_masses, fnc=f_b_measure):
    '''
    Calculate mean pairwise segmentation F-Measure.
    
    .. seealso:: :func:`f_b_measure`
    .. seealso:: :func:`recall`
    .. seealso:: :func:`precision`
    .. seealso:: :func:`segeval.compute_pairwise`
    
    :param dataset_masses: Segmentation mass dataset (including multiple \
                           codings).
    :type dataset_masses: dict
        
    :returns: Mean, standard deviation, variance, and standard error of a \
        segmentation metric.
    :rtype: :class:`decimal.Decimal`, :class:`decimal.Decimal`, \
        :class:`decimal.Decimal`, :class:`decimal.Decimal`
    '''
    return compute_pairwise(dataset_masses, fnc, permuted=False)


OUTPUT_NAME_F = 'Pairwise Mean F_beta Measure'
OUTPUT_NAME_R = 'Pairwise Mean Recall'
OUTPUT_NAME_P = 'Pairwise Mean Precision'
SHORT_NAME_F  = 'F_%s'
SHORT_NAME_P  = 'P'
SHORT_NAME_R  = 'R'


def values_f_b_measure(dataset_masses, beta=DEFAULT_BETA):
    '''
    Produces a TSV for this metric
    '''
    # Define a fnc to retrieve F_Beta-Measure values
    def wrapper_f(hypothesis_masses, reference_masses, return_parts=False):
        '''
        Wrapper to provide parameters.
        '''
        return f_b_measure(hypothesis_masses, reference_masses, beta,
                           return_parts=return_parts)
    # Create header
    header = list(['coder1', 'coder2', SHORT_NAME_F % str(beta), SHORT_NAME_P, \
                   SHORT_NAME_R, 'TP', 'FP', 'FN', 'TN'])
    # Calculate values
    values_f = compute_pairwise_values(dataset_masses, wrapper_f,
                                       permuted=False)
    values_p = compute_pairwise_values(dataset_masses, precision,
                                       permuted=False)
    values_r = compute_pairwise_values(dataset_masses, recall,
                                       permuted=False)
    values_cf = compute_pairwise_values(dataset_masses, wrapper_f,
                                        permuted=False, return_parts=True)
    # Combine into one table
    combined_values = dict()
    for label in values_f.keys():
        row = list()
        row.append(values_f[label])
        row.append(values_p[label])
        row.append(values_r[label])
        row.extend(values_cf[label])
        combined_values[label] = row
    # Return
    return create_tsv_rows(header, combined_values)


def parse(args):
    '''
    Parse this module's metric arguments and perform requested actions.
    '''
    output = None
    values = load_file(args)[0]
    beta = 1
    if 'beta' in args and args['beta'] != 1:
        beta = args['beta']
    # Is a TSV requested?
    if args['output'] != None:
        # Create a TSV
        output_file = args['output'][0]
        header, rows = values_f_b_measure(values, beta)
        write_tsv(output_file, header, rows)
    else:
        # Create a string to output
        subparser = args['subparser_name']
        if subparser == 'f':
            mean, std, var, stderr = pairwise_f_b_measure(values, beta)
            name = SHORT_NAME_F % str(beta)
        elif subparser == 'r':
            mean, std, var, stderr = pairwise_ml_measure(values, fnc=recall)
            name = SHORT_NAME_R
        elif subparser == 'p':
            mean, std, var, stderr = pairwise_ml_measure(values, fnc=precision)
            name = SHORT_NAME_P
        output = render_mean_values(name, mean, std, var, stderr)
    # Return
    return output


def parser_beta_support(parser):
    '''
    Add support for the "beta" parameter to allow for F_beta Measure calculation
    
    :param parser: Argument parser
    :type parser: argparse.ArgumentParser
    '''
    parser.add_argument('-b', '--beta',
                        type=float,
                        default=1.0,
                        help='Beta, the ratio of recall to precision in '+\
                            'F_beta measure; default is 1 (equal weight), '+\
                            '2 weights recall higher than precision, and '+\
                            '0.5 weights precision more than recall.')


def create_parser(subparsers):
    '''
    Setup a command line parser for this module's metric.
    '''
    from ..data import parser_add_file_support
    parser_f = subparsers.add_parser('f',
                                   help=OUTPUT_NAME_F)
    parser_add_file_support(parser_f)
    parser_f.set_defaults(func=parse)
    parser_beta_support(parser_f)
    parser = subparsers.add_parser('r',
                                   help=OUTPUT_NAME_R)
    parser_add_file_support(parser)
    parser.set_defaults(func=parse)

    parser = subparsers.add_parser('p',
                                   help=OUTPUT_NAME_P)
    parser_add_file_support(parser)
    parser.set_defaults(func=parse)




'''
Boundary edit distance for one boundary type [FournierInkpen2012]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from .MultipleBoundaryDistance import boundary_string_from_masses, \
    set_errors_transpositions_n_edits


def linear_edit_distance(segment_masses_a, segment_masses_b, n):
    '''
    Compute edit distance for linear single boundary type segmentations.
    
    :param segment_masses_a:  Segmentation masses.
    :param segment_masses_b:  Segmentation masses.
    :param n:                 The maximum number of PBs that boundaries can \
                                  span to be considered transpositions (n<2 \
                                  means no transpositions)
    :type segment_masses_a: list
    :type segment_masses_b: list
    :type beta: float
    '''
    # pylint: disable=C0103
    
    # Convert
    string_a = boundary_string_from_masses(segment_masses_a)
    string_b = boundary_string_from_masses(segment_masses_b)
    
    # There exist no boundaries at the beginning and end of linear single
    # boundary segmentations (that can differ)
    string_a[0]  = []
    string_a[-1] = []
    string_b[0]  = []
    string_b[-1] = []
    
    # Compute edit distance
    return set_errors_transpositions_n_edits(string_a, string_b, [1], n)



'''
Machine learning metric package.  This package a variety of traditional machine
learning metrics that have been adapted for use in segmentation, including:

* F-measure; and
* Percentage agreement.

.. warning:: These are provided for comparison, but are not recommended for \
    segmentation evaluation.  Instead, use  the segmentation similarity
    metric [FournierInkpen2012]_ implemented in
    :func:`segeval.similarity.SegmentationSimilarity.similarity` and the
    associated inter-coder agreement coefficients in
    :mod:`segeval.agreement`.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal


def load_tests(loader, tests, pattern):
    '''
    A ``load_tests()`` function utilizing the default loader
    :func:`segeval.Utils.default_load_tests`.
    
    .. seealso:: The `load_tests protocol <http://docs.python.org/library/\
    unittest.html#load-tests-protocol>`_.
    '''
    #pylint: disable=W0613
    from ..Utils import default_load_tests
    return default_load_tests(__file__, loader, tests)


def precision(tp, fp):
    '''
    Calculate precision.
    
    .. math::
        \\text{Precision} = \\frac{TP}{TP + FP}
    
    :param tp: Number of true positives
    :param fp: Number of false positives
    
    :returns: Precision.
    :rtype: :class:`decimal.Decimal`
    '''
    # pylint: disable=C0103
    if tp == 0:
        return Decimal(0)
    else:
        return Decimal(tp) / Decimal(tp + fp)
    
    
def recall(tp, fn):
    '''
    Calculate recall.
    
    .. math::
        \\text{Recall} = \\frac{TP}{TP + FN}
    
    Arguments:
    :param tp: Number of true positives
    :param fn: Number of false negatives
    
    :returns: Recall.
    :rtype: :class:`decimal.Decimal`
    '''
    # pylint: disable=C0103
    if tp == 0:
        return Decimal(0)
    else:
        return Decimal(tp) / Decimal(tp + fn)


def fmeasure(tp, fp, fn, beta=1.0):
    '''
    Calculate F-measure, also known as F-score.
    
    .. math::
        \\text{F}_{\\beta}\\text{-measure} = \\frac{(1 + \\beta^2) \\cdot TP}\
        {(1 + \\beta^2) \\cdot TP + \\beta^2 \\cdot FN + FP}
    
    :param tp: Number of true positives.
    :type tp: int
    :param fp: Number of false positives.
    :type fp: int
    :param fn: Number of false negatives.
    :type fn: int
    :param beta: Scales how precision and recall are averaged.
    :type beta: double
    
    :returns: F-measure.
    :rtype: :class:`decimal.Decimal`
    '''
    # pylint: disable=C0103
    if tp == 0 and fp == 0 and fn == 0:
        return Decimal('0')
    else:
        # Convert to Decimal
        tp   = Decimal(tp)
        fp   = Decimal(fp)
        fn   = Decimal(fn)
        beta = Decimal(str(beta))
        # Calculate terms
        beta2   = beta ** 2
        beta2_1 = Decimal('1.0') + beta2
        numerator   = beta2_1 * tp
        denomenator = (beta2_1 * tp) + (beta2 * fn) + fp
        # Perform division
        return numerator / denomenator


def confusionmatrix(tp, fp, fn, tn=None):
    '''
    Creates 2D tuples representing a confusion matrix.
    
    :param tp: Number of true positives.
    :param fp: Number of false positives.
    :param fn: Number of false negatives.
    :param tn: Number of true negatives.
    :type tp: int
    :type fp: int
    :type fn: int
    :type tn: int
    
    :returns: 2D tuples containing a confusion matrix.
    :rtype: :func:`tuple` containing a pair of :func:`tuple` objects, each \
        containing two :func:`int` values.
    '''
    # pylint: disable=C0103
    tp = Decimal(tp)
    fp = Decimal(fp)
    fn = Decimal(fn)
    return ( (tp, fp), (fn, tn) )


def cf_tostring(tp, fp, fn, tn=None):
    '''
    Creates a string representation of a confusion matrix.
    
    :param tp: Number of true positives.
    :param fp: Number of false positives.
    :param fn: Number of false negatives.
    :param tn: Number of true negatives.
    :type tp: int
    :type fp: int
    :type fn: int
    :type tn: int
    
    :returns: String representation of a confusion matrix.
    :rtype: :func:`str`
    '''
    # pylint: disable=C0103
    return '[%(tp)i \t %(fp)i\n %(fn)i \t %(tn)i]' % \
        {'tp': tp, 'fp' : fp, 'fn' : fn, 'tn' : tn}


def prfcf(tp, fp, fn, tn=None, beta=1.0):
    '''
    Calculates precision, recall, F-Score, and creates a confusion matrix.
    
    :param tp:   Number of true positives.
    :param fp:   Number of false positives.
    :param fn:   Number of false negatives.
    :param beta: Scales how precision and recall are averaged.
    :type tp:    int
    :type fp:    int
    :type fn:    int
    :type beta:  float
    
    :returns: Precision, recall, F-measure, and a confusion matrix.
    :rtype: :class:`decimal.Decimal`, :class:`decimal.Decimal`, \
        :class:`decimal.Decimal`, \
        :func:`tuple` containing a pair of :func:`tuple` objects, each \
        containing two :func:`int` values.
    '''
    # pylint: disable=C0103
    return precision(tp, fp), recall(tp, fn), fmeasure(tp, fp, fn, beta), \
           confusionmatrix(tp, fp, fn, tn)


def find_boundary_position_freqs(masses_set):
    '''
    Converts a list of segmentation mass sets into a dict of boundary positions,
    and the frequency of the whether boundaries were chosen at that location or,
    or not,
    
    :param masses_set: List of segmentation masses
    :type masses_set: list
    
    :returns: :func:`dict` of boundary position frequencies.
    :rtype: :class:`decimal.Decimal`
    '''
    seg_positions = dict()
    for masses in masses_set:
        # Iterate over boundary positions that precede the index, excluding the
        # first position so that the start is not counted as a boundary
        for i in xrange(1, len(masses)):
            position = sum(masses[0:i])
            try:
                seg_positions[position] += 1
            except KeyError:
                seg_positions[position] = 1
    return seg_positions




'''
Tests the boundary distance and string methods functionality.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from .SingleBoundaryDistance import linear_edit_distance


class TestBoundaryDistance(unittest.TestCase):
    '''
    Test single boundary type edit distance.
    '''
    # pylint: disable=R0904,C0324,C0103
    
    SKIP = False
    
    def test_st_2_basic_cases(self):
        '''
        Substitution and transpsition (n=2) basic cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Transposition
        mass_a = [2,2]
        mass_b = [3,1]
        d,t,s = linear_edit_distance(mass_a, mass_b, 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        
        d,t,s = linear_edit_distance(mass_a, mass_b, 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        
        # Add/sub/del
        mass_a = [4,1]
        mass_b = [1,4]
        d,t,s = linear_edit_distance(mass_a, mass_b, 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        
        mass_a = [4]
        mass_b = [4]
        d,t,s = linear_edit_distance(mass_a, mass_b, 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (0,0,0))


    def test_st_2_edge_cases(self):
        '''
        Substitution and transpsition (n=2) edge cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        d,t,s = linear_edit_distance([], [], 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (0,0,0))


    def test_st_2_complex_cases(self):
        '''
        Substitution and transpsition (n=2) complex cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        mass_a = [1,2,2,4]
        mass_b = [2,2,5]
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance(mass_a, mass_b, 2)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,2,1))


    def test_nested_transp_case(self):
        '''
        Nested transpsition case.
        '''
        if TestBoundaryDistance.SKIP:
            return
        mass_a = [1,1,7]
        mass_b = [5,1,3]
        d,t,s = linear_edit_distance(mass_a, mass_b, 7)[0:3]
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        d,t,s = linear_edit_distance(mass_a, mass_b, 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,1,2))
    
    
    def test_st_1_basic_cases(self):
        '''
        Substitution and transpsition (n=1) basic cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Transposition
        d,t,s = linear_edit_distance([2,2],
                                     [3,1], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        d,t,s = linear_edit_distance([3,1],
                                     [2,2], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        # Add/sub/del
        d,t,s = linear_edit_distance([5,1],
                                     [1,5], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        d,t,s = linear_edit_distance([4],
                                     [4], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (0,0,0))


    def test_st_1_edge_cases(self):
        '''
        Substitution and transpsition (n=1) edge cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        d,t,s = linear_edit_distance([],
                                     [], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (0,0,0))


    def test_st_1_complex_cases(self):
        '''
        Substitution and transpsition (n=1) complex cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,5], 1)[0:3]
        self.assertEqual((d,len(t),len(s)), (5,0,5))
    
    
    def test_st_3_basic_cases(self):
        '''
        Substitution and transpsition (n=3) basic cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Transposition
        d,t,s = linear_edit_distance([1,3],
                                     [3,1], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        # Transposition
        d,t,s = linear_edit_distance([2,2],
                                     [3,1], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (1,1,0))
        # Add/sub/del
        d,t,s = linear_edit_distance([5,1],
                                     [1,5], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (2,0,2))
        d,t,s = linear_edit_distance([4],
                                     [4], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (0,0,0))


    def test_st_3_complex_cases(self):
        '''
        Substitution and transpsition (n=3) complex cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,5], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,2,1))
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,3,2], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
    
    
    def test_st_3_mixup_cases(self):
        '''
        Substitution and transpsition (n=3) mixup (tricky) cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,3],
                                     [2,2,2,2], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,3,2], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,5],
                                     [2,2,4,2], 3)[0:3]
        self.assertEqual((d,len(t),len(s)), (4,2,2))
        
    def test_st_4_mixup_cases(self):
        '''
        Substitution and transpsition (n=4) mixup (tricky) cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,3],
                                     [2,2,2,2], 4)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,3,2], 4)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,5],
                                     [2,2,4,2], 4)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,6],
                                     [2,2,5,2], 4)[0:3]
        self.assertEqual((d,len(t),len(s)), (4,2,2))
        
    def test_st_5_mixup_cases(self):
        '''
        Substitution and transpsition (n=5) mixup (tricky) cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,3],
                                     [2,2,2,2], 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,3,2], 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,5],
                                     [2,2,4,2], 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,6],
                                     [2,2,5,2], 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,7],
                                     [2,2,6,2], 5)[0:3]
        self.assertEqual((d,len(t),len(s)), (4,2,2))
        
    def test_st_6_mixup_cases(self):
        '''
        Substitution and transpsition (n=6) mixup (tricky) cases.
        '''
        if TestBoundaryDistance.SKIP:
            return
        # Combination of add/sub/del and transposition
        d,t,s = linear_edit_distance([1,2,2,3],
                                     [2,2,2,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,4],
                                     [2,2,3,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,5],
                                     [2,2,4,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,6],
                                     [2,2,5,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,7],
                                     [2,2,6,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (3,3,0))
        d,t,s = linear_edit_distance([1,2,2,8],
                                     [2,2,7,2], 6)[0:3]
        self.assertEqual((d,len(t),len(s)), (4,2,2))




'''
Inter-coder agreement metrics.  This module provides inter-coder agreement
metrics adapted for segmentation (read [FournierInkpen2012]_) that make use of
segmentation similarity [FournierInkpen2012]_, including:

* Kappa [Cohen1960]_, [DaviesFleiss1982]_ and
* Pi [Scott1955]_, [Fleiss1971]_

To discuss measures of inter-coder (or inter-annotator) agreement, we use a
modified version of [ArtsteinPoesio2008]_'s terminology, as presented in
[FournierInkpen2012]_, where the set of:

* *Items* is :math:`\{i|i \in I\}` with cardinality **i**;
* *Categories* is :math:`\{k|k \in K\}` with cardinality **k**;
* *Coders* is :math:`\{c|c \in C\}` with cardinality **c**;
* *Segmentations* of an item :math:`i` by a coder :math:`c` is 
  :math:`\{s|s \in S\}`, where when :math:`s_{ic}`  is specified with only one
  subscript, it denotes :math:`s_{c}`, for all relevant items (:math:`i`); and
* *Types* of segmentation boundaries is :math:`\{t|t \in T\}` with
  cardinality **t**.

For segmentation, we assume that the set of categories is the
presence of a segmentation boundary :math:`\\text{seg}` of type
:math:`t` at a position in a segmentation [FournierInkpen2012]_:

.. math::
    K = \\{\\text{seg}_t | t \in T\\}

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import math
from decimal import Decimal
from ..similarity.SegmentationSimilarity import similarity as similarity_linear


def load_tests(loader, tests, pattern):
    '''
    A ``load_tests()`` function utilizing the default loader
    :func:`segeval.Utils.default_load_tests`.
    
    .. seealso:: The `load_tests protocol <http://docs.python.org/library/\
    unittest.html#load-tests-protocol>`_.
    '''
    #pylint: disable=W0613
    from ..Utils import default_load_tests
    return default_load_tests(__file__, loader, tests)


def choose(n, k):
    '''
    Calculates a number of combinations (i.e., 4 choose 2).
    
    .. math::
        C(n,k) = \\frac{n!}{k!(n-k)!}
    
    :param n: Number of elements to choose from.
    :param k: Number of elements to chose.
    
    :returns: Number of combinations.
    :rtype: :func:`int`
    '''
    # pylint: disable=C0103
    numerator   = math.factorial(n)
    denominator = (math.factorial(k) * math.factorial(n-k))
    return numerator / denominator


def permute(n, k):
    '''
    Calculates a number of permutations (i.e., 4 permute 2).
    
    .. math::
        P(n,k) = \\frac{n!}{(n-k)!}
    
    :param n: Number of elements to permute from.
    :param k: Number of elements to permute.
    
    :returns: Number of permutations.
    :rtype: :func:`int`
    '''
    # pylint: disable=C0103
    numerator   = math.factorial(n)
    denominator = (math.factorial(n-k))
    return numerator / denominator


def actual_agreement(items_masses):
    '''
    Calculate actual (i.e., observed or :math:`\\text{A}_a`), segmentation
    agreement without accounting for chance, using [ArtsteinPoesio2008]_'s
    formulation as adapted in [FournierInkpen2012]_:
    
    .. math::
        \\text{A}_a = \\frac{
            \sum_{i \in I} \\text{mass}(i) \cdot \\text{S}(s_{i1},s_{i2})
        }{
            \sum_{i \in I} \\text{mass}(i) 
        }
        
    Or, for more than two coders:
    
    .. math::
        \\text{A}_a = \\frac{1}{{\\textbf{c} \\choose 2}}
        \sum^{\\textbf{c}-1}_{m=1}
        \sum^{\\textbf{c}}_{n=m+1}
        \\frac{
            \sum_{i \in I} \\text{mass}(i) \cdot \\text{S}(s_{im},s_{in})
        }{
            \sum_{i \in I} \\big( \\text{mass}(i) - 1 \\big)
        }
        
    Where :math:`\\text{S}(s_{i1},s_{i2})` is defined in 
    :func:`segeval.similarity.SegmentationSimilarity.similarity`.
    
    :param items_masses: Segmentation masses for a collection of items where \
                        each item is multiply coded (all coders code all items).
    :type items_masses: dict
    
    :returns: Potential boundaries unmoved, all potential boundaries, and the \
              boundaries per coder.
    :rtype: :func:`list`, :func:`list`, :func:`dict`
    
    An example of the dictionary structure if the ``items_masses``
    parameter is::
    
            items_masses = {
                'item1' : {
                    'coder1' : [5],
                    'coder2' : [2,3],
                    'coder2' : [1,1,3]
                },
                'item2' : {
                    'coder1' : [8],
                    'coder2' : [4,4],
                    'coder2' : [2,2,4]
                }
            }
    
    Other real and contrived examples can be found in 
    :mod:`segeval.data.Samples`.
    
    '''
    # pylint: disable=C0103
    all_pbs_unedited = list()
    all_pbs          = list()
    coders_boundaries = dict()
    coders = items_masses.values()[0].keys()
    # FOr each permutation of coders
    for m in range(0, len(coders) - 1):
        for n in range(m+1, len(coders)):
            for item in items_masses.keys():
                segs_a = items_masses[item][coders[m]]
                segs_b = items_masses[item][coders[n]]
                pbs_unedited, total_pbs = \
                    similarity_linear(segs_a, segs_b, return_parts=True)[0:2]
                all_pbs_unedited.append(pbs_unedited)
                all_pbs.append(total_pbs)
                # Create in dicts if not present
                if coders[m] not in coders_boundaries:
                    coders_boundaries[coders[m]] = list()
                if coders[n] not in coders_boundaries:
                    coders_boundaries[coders[n]] = list()
                # Add per-coder values to dicts
                coders_boundaries[coders[m]].append(
                    [Decimal(len(segs_a)),
                     total_pbs])
                coders_boundaries[coders[n]].append(
                    [Decimal(len(segs_b)),
                     total_pbs])
    return all_pbs_unedited, all_pbs, coders_boundaries




'''
Tests the data i/o functions and package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import os
import unittest
from . import load_nested_folders_dict, input_linear_mass_tsv, \
    input_linear_positions_tsv, input_linear_mass_json, FILETYPE_JSON
from .Samples import HEARST_1997


class TestData(unittest.TestCase):
    '''
    Test data i/o functions and package.
    '''
    #pylint: disable=R0904,C0103
    
    test_data_dir = os.path.split(__file__)[0]

    def test_load_nested_folders_dict(self):
        '''
        Test nested folder dict construction.
        '''
        data = load_nested_folders_dict(os.path.join(self.test_data_dir, '..'),
                                        FILETYPE_JSON)
        self.assertEqual(data['data']['hearst1997'], HEARST_1997)
    
    def test_input_linear_mass_tsv(self):
        '''
        Test mass TSV file input.
        '''
        tsv_file = os.path.join(self.test_data_dir, 'hearst1997.tsv')
        data = input_linear_mass_tsv(tsv_file)
        self.assertEqual(data, HEARST_1997)
      
    def test_input_linear_positions_tsv(self):
        '''
        Test position TSV file input.
        '''
        tsv_file = os.path.join(self.test_data_dir, 'hearst1997_positions.csv')
        data = input_linear_positions_tsv(tsv_file, delimiter=',')
        self.assertEqual(data, HEARST_1997)
    
    def test_input_linear_mass_json(self):
        '''
        Test mass JSON file input.
        '''
        json_file = os.path.join(self.test_data_dir, 'hearst1997.json')
        data = input_linear_mass_json(json_file)
        self.assertEqual(data, HEARST_1997)




'''
Tests the machine learning (ML) statistics functions, and ml package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from decimal import Decimal
from . import precision, recall, fmeasure, confusionmatrix, prfcf


class TestML(unittest.TestCase):
    '''
    Machine-learning metric tests.
    '''
    #pylint: disable=R0904,C0103
    
    def test_precision(self):
        '''
        Test precision.
        '''
        tp = 1
        fp = 1
        self.assertEqual(precision(tp, fp), Decimal('0.5'))
        tp = 1
        fp = 3
        self.assertEqual(precision(tp, fp), Decimal('0.25'))
        tp = 6
        fp = 2
        self.assertEqual(precision(tp, fp), Decimal('0.75'))
        tp = 0
        fp = 2
        self.assertEqual(precision(tp, fp), Decimal('0'))
        tp = 2
        fp = 0
        self.assertEqual(precision(tp, fp), Decimal('1'))
        tp = 0
        fp = 0
        self.assertEqual(precision(tp, fp), Decimal('0'))
        
    def test_recall(self):
        '''
        Test recall.
        '''
        tp = 1
        fn = 1
        self.assertEqual(recall(tp, fn), Decimal('0.5'))
        tp = 1
        fn = 3
        self.assertEqual(recall(tp, fn), Decimal('0.25'))
        tp = 6
        fn = 2
        self.assertEqual(recall(tp, fn), Decimal('0.75'))
        tp = 0
        fn = 2
        self.assertEqual(recall(tp, fn), Decimal('0'))
        tp = 2
        fn = 0
        self.assertEqual(recall(tp, fn), Decimal('1'))
        tp = 0
        fn = 0
        self.assertEqual(recall(tp, fn), Decimal('0'))
        
    def test_f1(self):
        '''
        Test F-Score with a beta of 1.0, 0.5, or 2.0.
        '''
        tp = 2
        fp = 1
        fn = 1
        beta = 1.0
        self.assertEqual(fmeasure(tp, fp, fn, beta),
                         Decimal('4') / Decimal('6'))
        tp = 1
        fp = 3
        fn = 1
        beta = 1.0
        self.assertEqual(fmeasure(tp, fp, fn, beta),
                         Decimal('1') / Decimal('3'))
        beta = 0.5
        self.assertAlmostEqual(fmeasure(tp, fp, fn, beta),
                               Decimal('0.277777777'))
        beta = 2.0
        self.assertAlmostEqual(fmeasure(tp, fp, fn, beta),
                               Decimal('0.416666666'))
        tp = 0
        fp = 0
        fn = 0
        beta = 1.0
        self.assertEqual(fmeasure(tp, fp, fn, beta), Decimal('0'))
        
    def test_confusion_matrix(self):
        '''
        Tests confusion matrix calculation.
        '''
        tp = 1
        fp = 2
        fn = 3
        tn = None
        
        cf = ( (tp, fp), (fn, tn) )
        self.assertEqual(confusionmatrix(tp, fp, fn, tn), cf)
    
    def test_prfcf(self):
        '''
        Tests precision, recall, F1, and CF calculation.
        '''
        tp = 2
        fp = 4
        fn = 8
        tn = None
        
        p = Decimal('1') / Decimal('3')
        r = Decimal('0.2')
        f = Decimal('0.25')
        
        cf = ( (tp, fp), (fn, tn) )
        self.assertEqual(prfcf(tp, fp, fn, tn), (p, r, f, cf))




'''
Provides a segmentation version of the  percentage agreement metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal
from . import find_boundary_position_freqs
from .. import compute_pairwise, compute_pairwise_values, create_tsv_rows
from ..data import load_file
from ..data.TSV import write_tsv
from ..data.Display import render_mean_values


def percentage(hypothesis_masses, reference_masses):
    '''
    Calculates the percentage agreement between a hypothesis, and reference
    segmentation.
    
    :param hypothesis_masses: Hypothesis segmentation masses.
    :param reference_masses: Reference segmentation masses.
    :param beta: Scales how precision and recall are averaged.
    :type hypothesis_masses: list
    :type reference_masses: list
    
    :returns: F-measure.
    :rtype: :class:`decimal.Decimal`
    '''
    positions_hyp = find_boundary_position_freqs([hypothesis_masses])
    positions_ref = find_boundary_position_freqs([reference_masses])
    agree    = Decimal('0')
    disagree = Decimal('0')
    for pos in positions_ref.keys():
        if pos in positions_hyp:
            agree += 1
        else:
            disagree += 1
    for pos in positions_hyp.keys():
        if pos not in positions_ref:
            disagree += 1
    return agree / (agree + disagree)


def pairwise_percentage(dataset_masses):
    '''
    Calculate mean pairwise segmentation percentage correctness.
    
    .. seealso:: :func:`percentage`
    .. seealso:: :func:`segeval.compute_pairwise`
    
    :param dataset_masses: Segmentation mass dataset (including multiple \
                           codings).
    :type dataset_masses: dict
        
    :returns: Mean, standard deviation, variance, and standard error of a \
        segmentation metric.
    :rtype: :class:`decimal.Decimal`, :class:`decimal.Decimal`, \
        :class:`decimal.Decimal`, :class:`decimal.Decimal`
    '''
    return compute_pairwise(dataset_masses, percentage, permuted=False)


OUTPUT_NAME = 'Pairwise Mean Percentage'
SHORT_NAME  = 'Pr'


def values_percentage(dataset_masses):
    '''
    Produces a TSV for this metric
    '''
    header = list(['coder1', 'coder2', SHORT_NAME])
    values = compute_pairwise_values(dataset_masses, percentage)
    return create_tsv_rows(header, values)


def parse(args):
    '''
    Parse this module's metric arguments and perform requested actions.
    '''
    output = None
    values = load_file(args)[0]
    # Is a TSV requested?
    if args['output'] != None:
        # Create a TSV
        output_file = args['output'][0]
        header, rows = values_percentage(values,)
        write_tsv(output_file, header, rows)
    else:
        # Create a string to output
        mean, std, var, stderr = pairwise_percentage(values)
        name = SHORT_NAME
        output = render_mean_values(name, mean, std, var, stderr)
    return output


def create_parser(subparsers):
    '''
    Setup a command line parser for this module's metric.
    '''
    from ..data import parser_add_file_support
    parser = subparsers.add_parser('pr',
                                   help=OUTPUT_NAME)
    parser_add_file_support(parser)
    parser.set_defaults(func=parse)




'''
Console interface to the overall segeval package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import argparse
import sys
from segeval.agreement.Pi import create_parser as create_parser_pi
from segeval.agreement.Kappa import create_parser as create_parser_kappa
from segeval.agreement.Bias import create_parser as create_parser_bias
from segeval.ml.FbMeasure import create_parser as create_parser_fmeasure
from segeval.ml.Percentage import create_parser as create_parser_percentage
from segeval.similarity.SegmentationSimilarity import create_parser as \
    create_parser_similarity
from segeval.window.Pk import create_parser as create_parser_pk
from segeval.window.WindowDiff import create_parser as create_parser_windowdiff
from segeval.window.WinPR import create_parser as create_parser_winpr


def main(argv=None):
    '''
    Main method for command line parsing and actions.
    '''
    parser = argparse.ArgumentParser(prog='segeval',
                                     description='A discourse segmentation '+\
                                            'evaluation utility.')
    
    subparsers = parser.add_subparsers(title='metric', 
                                       description='Calculates a specified '+\
                                            'segmentation evaluation metric '+\
                                            'upon provided data',
                                       help='Available metrics',
                                       dest='subparser_name')
    
    create_parser_pi(subparsers)
    create_parser_kappa(subparsers)
    create_parser_bias(subparsers)
    create_parser_fmeasure(subparsers)
    create_parser_percentage(subparsers)
    create_parser_similarity(subparsers)
    create_parser_pk(subparsers)
    create_parser_windowdiff(subparsers)
    create_parser_winpr(subparsers)
    # Parse arguments
    args = None
    if argv:
        args = parser.parse_args(argv)
    else:
        args = parser.parse_args()
    
    return args.func(vars(args))


if __name__ == '__main__':
    # pylint: disable=C0103
    output = main()
    if output != None:
        print output
    sys.exit(0)




'''
This module contains implementations of Scott's and Fleiss' Pi, adapted in
[FournierInkpen2012]_ for segmentation (in as similar a manner as [Hearst1997]_)
using the formulation of Kappa provided in [ArtsteinPoesio2008]_.

Pi's general form could be described, as it is in [ArtsteinPoesio2008]_, in 
terms of actual agreement (:math:`\\text{A}_a`) and expected agreement 
(:math:`\\text{A}_e`) as:

.. math::
    \pi,\pi^* = \\frac{\\text{A}_a-\\text{A}_e}{1 \
    - \\text{A}_e}

:math:`\pi` represents Scott's Pi (for 2 coders), whereas :math:`\pi^*`
represents its generalization to more than 2 coders.  Each metric calculates
:math:`\\text{A}_a` using :func:`segeval.agreement.actual_agreement` and
only varies the calculation of :math:`\\text{A}_e`.


.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal
from . import actual_agreement
from .. import compute_mean, compute_mean_values, create_tsv_rows
from ..data import load_file
from ..data.TSV import write_tsv
from ..data.Display import render_value, render_mean_values


def scotts_pi(items_masses, return_parts=False):
    '''
    Calculates Scott's Pi, originally proposed in [Scott1955]_, for
    segmentations.  Adapted in [FournierInkpen2012]_ from the formulations
    provided in [Hearst1997]_ and [ArtsteinPoesio2008]_'s formulation for
    expected agreement:
    
    .. math::
        \\text{A}^\pi_e = \sum_{k \in K} \\big(\\text{P}^\pi_e(k)\\big)^2
    
    .. math::
        \\text{P}^\pi_e(\\text{seg}_t) = 
        \\frac{
            \sum_{c \in C}\sum_{i \in I}|\\text{boundaries}(t, s_{ic})|
        }{
            \\textbf{c} \cdot \sum_{i \in I} \\big( \\text{mass}(i) - 1 \\big)
        }
    
    :param items_masses: Segmentation masses for a collection of items where \
                        each item is multiply coded (all coders code all items).
    :param return_parts: If true, return the numerator and denominator
    :type item_masses:  dict
    :type return_parts: bool
    
    :returns: Scott's Pi
    :rtype: :class:`decimal.Decimal`
    
    .. seealso:: :func:`segeval.agreement.actual_agreement` for an example of\
     ``items_masses``.
    
    .. note:: Applicable for only 2 coders.
    '''
    # Check that there are no more than 2 coders
    if len([True for coder_segs in items_masses.values() \
            if len(coder_segs.keys()) > 2]) > 0:
        raise Exception('Unequal number of items specified.')
    # Check that there are an identical number of items
    num_items = len(items_masses.values()[0].keys())
    if len([True for coder_segs in items_masses.values() \
            if len(coder_segs.values()) != num_items]) > 0:
        raise Exception('Unequal number of items contained.')
    # Return
    return fleiss_pi(items_masses, return_parts)


def fleiss_pi(items_masses, return_parts=False):
    '''
    Calculates Fleiss' Pi (or multi-Pi), originally proposed in [Fleiss1971]_,
    for segmentations (and described in [SiegelCastellan1988]_ as K).
    Adapted from the formulations
    provided in [Hearst1997]_ (p. 53) and [ArtsteinPoesio2008]_'s formulation
    for expected agreement:
    
    .. math::
        \\text{A}^{\pi^*}_e = \sum_{k \in K} \\big(\\text{P}^\pi_e(k)\\big)^2
    
    :param items_masses: Segmentation masses for a collection of items where \
                        each item is multiply coded (all coders code all items).
    :param return_parts: If true, return the numerator and denominator.
    :type items_masses:  dict
    :type return_parts: bool
    
    :returns: Fleiss's Pi
    :rtype: :class:`decimal.Decimal`
    
    .. seealso:: :func:`segeval.agreement.actual_agreement` for an example of\
     ``items_masses``.
    
    .. note:: Applicable for more than 2 coders.
    '''
    # pylint: disable=C0103,R0914
    # Check that there are an equal number of items for each coder
    num_items = len(items_masses.values()[0].keys())
    if len([True for coder_segs in items_masses.values() \
            if len(coder_segs.values()) != num_items]) > 0:
        raise Exception('Unequal number of items contained.')
    # Initialize totals
    unmoved_masses, total_masses, coders_boundaries_totalboundaries = \
        actual_agreement(items_masses)
    # Calculate Aa
    A_a = Decimal(sum(unmoved_masses)) / sum(total_masses)
    # Calculate Ae
    p_e_segs = list()
    for boundaries_info in coders_boundaries_totalboundaries.values():
        for item in boundaries_info:
            boundaries, total_boundaries = item
            p_e_seg = boundaries / total_boundaries
            p_e_segs.append(p_e_seg)
    # Calculate P_e_seg
    P_e_seg = Decimal(sum(p_e_segs)) / len(p_e_segs)
    A_e = (P_e_seg ** 2)
    # Calculate pi
    pi = (A_a - A_e) / (Decimal('1.0') - A_e)
    # Return
    if return_parts:
        return A_a, A_e
    else:
        return pi


def mean_fleiss_pi(dataset_masses):
    '''
    Calculate mean segmentation Fleiss' Pi.
    
    .. seealso:: :func:`fleiss_pi`, :func:`segeval.compute_mean`
    
    :param dataset_masses: Segmentation mass dataset (including multiple \
                           codings).
    :type dataset_masses: dict
    
    :returns: Mean, standard deviation, and variance.
    :rtype: :class:`decimal.Decimal`, :class:`decimal.Decimal`, :class:`decimal.Decimal`
    '''
    return compute_mean(dataset_masses, fleiss_pi)


OUTPUT_NAME     = 'S-based Fleiss\' Multi Pi'
SHORT_NAME      = 'Pi*_s'
SHORT_NAME_MEAN = 'Mean %s' % SHORT_NAME


def values_pi(dataset_masses):
    '''
    Produces a TSV for this metric
    '''
    header = list([SHORT_NAME])
    values = compute_mean_values(dataset_masses, fleiss_pi)
    return create_tsv_rows(header, values)


def parse(args):
    '''
    Parse this module's metric arguments and perform requested actions.
    '''
    output = None
    values, is_file = load_file(args)
    # Is a TSV requested?
    if args['output'] != None:
        # Create a TSV
        output_file = args['output'][0]
        header, rows = values_pi(values)
        write_tsv(output_file, header, rows)
    else:
        # Create a string to output
        if not args['output'] and is_file:
            # Render for one item
            output = render_value(SHORT_NAME, str(fleiss_pi(values)))
        else:
            # Render for one or more items
            mean, std, var, stderr = mean_fleiss_pi(values)
            output = render_mean_values(SHORT_NAME_MEAN, mean, std, var, stderr)
    # Return
    return output


def create_parser(subparsers):
    '''
    Setup a command line parser for this module's metric.
    '''
    from ..data import parser_add_file_support
    parser = subparsers.add_parser('pi',
                                   help=OUTPUT_NAME)
    parser_add_file_support(parser)
    parser.set_defaults(func=parse)




'''
Tests the WindowDiff evaluation metric.

@author: Chris Fournier
@contact: chris.m.fournier@gmail.com
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from decimal import Decimal
from ..window.Pk import pk, pairwise_pk
from ..data.Samples import KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, \
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT


class TestPk(unittest.TestCase):
    '''
    Test Pk.
    '''
    # pylint: disable=R0904

    def test_identical(self):
        '''
        Test whether identical segmentations produce 0.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b), 0.0)
        self.assertEqual(pk(a, b, one_minus=True), 1.0)

    def test_no_boundaries(self):
        '''
        Test whether no segments versus some segments produce 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(b, a),
                         Decimal('1.0'))
        self.assertEqual(pk(a, b),
                         Decimal('0.3636363636363636363636363636'))

    def test_all_boundaries(self):
        '''
        Test whether all segments versus some segments produces 7/11 = 0.636
        erroneous windows.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b),
                         Decimal('0.6363636363636363636363636364'))
        self.assertEqual(pk(b, a),
                         Decimal('0.6363636363636363636363636364'))

    def test_all_and_no_boundaries(self):
        '''
        Test whether all segments versus no segments produces 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        self.assertEqual(pk(a, b), 1.0)
        self.assertEqual(pk(b, a), 1.0)

    def test_translated_boundary(self):
        '''
        Test whether 2/3 total segments participate in mis-alignment produces
        0.182.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b),
                         Decimal('0.1818181818181818181818181818'))
        self.assertEqual(pk(b, a),
                         Decimal('0.1818181818181818181818181818'))
    
    def test_extra_boundary(self):
        '''
        Test whether 1/3 segments that are non-existent produces 0.091.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b),
                         Decimal('0.09090909090909090909090909091'))
        self.assertEqual(pk(b, a),
                         Decimal('0.09090909090909090909090909091'))
    
    def test_full_miss_and_misaligned(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.273. 
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b),
                         Decimal('0.2727272727272727272727272727'))
        self.assertEqual(pk(b, a),
                         Decimal('0.2727272727272727272727272727'))


class TestPairwisePkMeasure(unittest.TestCase):
    # pylint: disable=R0904
    '''
    Test pairwise Pk.
    '''
    
    def test_kazantseva2012_g5(self):
        '''
        Calculate permuted pairwise Pk on Group 5 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertEqual(pairwise_pk(KAZANTSEVA2012_G5,
                                     convert_from_masses=True),
                         (Decimal('0.3553005828239669303205984456'),
                          Decimal('0.1100176084609921563234047475'),
                          Decimal('0.01210387417147617290343989767'),
                          Decimal('0.01587967396513816764352290375')))
    
    def test_kazantseva2012_g2(self):
        '''
        Calculate mean permuted pairwise Pk on Group 2 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertEqual(pairwise_pk(KAZANTSEVA2012_G2,
                                     convert_from_masses=True),
                         (Decimal('0.2882256923776327507173609772'),
                          Decimal('0.1454395656787966169084191446'),
                          Decimal('0.02115266726483699483402909758'),
                          Decimal('0.01327675514600517730547602482')))
    
    def test_large_disagreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing large disagreement.
        '''
        self.assertEqual(pairwise_pk(LARGE_DISAGREEMENT,
                                     convert_from_masses=True),
                         (1.0,
                          0.0,
                          0.0,
                          0.0))
    
    def test_complete_agreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing complete agreement.
        '''
        self.assertEqual(pairwise_pk(COMPLETE_AGREEMENT,
                                     convert_from_masses=True),
                         (0.0,
                          0.0,
                          0.0,
                          0.0))




'''
Multiple-boundary edit distance.  Detailed in [FournierInkpen2012]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal


NO_BOUNDARY = '-'
BOUNDARY    = '|'


def boundary_string_from_masses(segment_masses):
    '''
    Creates a "boundary string", or sequence of boundary type sets.
    
    :param segment_masses: Segmentation masses.
    :type segment_masses:  list
    :returns: A sequence of boundary type sets
    :rtype: :func:`list` of :func:`set` onjects containing :func:`int` values.
    '''
    string = [set() for _ in xrange(0, sum(segment_masses) + 1)]
    
    pos = 0
    for mass in segment_masses:
        string[pos].add(1)
        string[pos + mass].add(1)
        pos += mass
    
    return [list(pb) for pb in string]



class SetError(object):
    '''
    Represents a set error at a particular potential boundary position.
    '''
    # pylint: disable=R0903
    
    def __init__(self, position, boundaries, boundary_types):
        '''
        Creates an object that represents a set error.
        
        :param position:      Position where the error occurs
        :param boundaries:    Number of boundaries involved.
        :param boundary_type: Boundary types involved.
        :type position:    int
        :type boundaries:  bool
        :type types:       bool
        '''
        # pylint: disable=C0103
        self.start          = position
        self.end            = position
        self.boundaries     = boundaries
        self.boundary_types = boundary_types
    
    def overlaps(self, other):
        '''
        Tests to see whether a transposition error overlaps with this set error.
        
        :param other: Other error object (usually a :func:`Transposition`)
        
        :returns: True if overlapping, False otherwise.
        :rtype: :func:`bool`
        '''
        overlap = False
        # Test bounds
        if other.boundary_type in self.boundary_types:
            if other.start >= self.start and other.start <= self.end:
                overlap = True
            elif other.end >= self.start and other.end <= self.end:
                overlap = True
            elif self.start >= other.start and self.start <= other.end:
                overlap = True
            elif self.end >= other.start and self.end <= other.end:
                overlap = True
        # Return
        return overlap
    
    def __str__(self):
        return '(%i,%i) b=%i t=%s' % (self.start, self.end,
                                        self.boundaries,
                                        ','.join(self.boundary_types))


def compare_sets(position, set_a, set_b):
    '''
    Calculate the number of substitution or addition/deletion operations
    required to edit one set to be the other.
    
    :param position: Position to calculate set error upon
    :param set_a:    First set to edit
    :param set_b:    Second set to edit
    :type position:  int
    :type set_a:     :func:`set` or :func:`list`
    :type set_b:     :func:`set` or :func:`list`
    
    :returns: A list of the substitutions and list of additions/deletions /
        performed
    :rtype: :func:`list`, :func:`list`
    '''
    # pylint: disable=W0141
    set_a, set_b = set(set_a), set(set_b)
    
    diff_ab = set_a.difference(set_b)
    diff_ba = set_b.difference(set_a)
    
    pairs = map(None, list(diff_ab), list(diff_ba))
    
    num_substitutions = min(len(diff_ab), len(diff_ba))
    num_add_or_del    = abs(len(diff_ab) - len(diff_ba))
    
    substitutions = list()
    add_or_del    = list()
    
    for pair in pairs:
        pair = list(pair)
        if None in pair:
            pair.remove(None)
            add_or_del.append(SetError(position, 1, pair))
        else:
            substitutions.append(SetError(position, 2, pair))
    
    if num_substitutions != len(substitutions):
        raise Exception('Incorrect substitution calculation')
    if num_add_or_del != len(add_or_del):
        raise Exception('Incorrect add_or_del calculation')
    
    return substitutions, add_or_del


def calculate_set_errors(string_a, string_b):
    '''
    Calculate set errors for two parallel boundary strings produced by
    :func:`boundary_string_from_masses`.
    
    .. seealso:: :func:`boundary_string_from_masses`
    
    :param string_a: Boundary string.
    :param string_b: Boundary string.
    :type string_a:  :func:`list` of :func:`set` onjects containing :func:`int`
    :type string_b:  :func:`list` of :func:`set` onjects containing :func:`int`
    :returns: Number of errors, and detailed information on errors as a list of\
        error objects
    :rtype: func:`int`, :func:`list`
    '''
    set_error_details = list()
    for i in range(0, len(string_a)):
        boundary_type_a_i = []
        boundary_type_b_i = []
        try:
            boundary_type_a_i = string_a[i]
        except KeyError:
            pass
        try:
            boundary_type_b_i = string_b[i]
        except KeyError:
            pass
        substitutions, add_or_del = compare_sets(i,
                                                 boundary_type_a_i,
                                                 boundary_type_b_i)
        set_error_details.extend(substitutions)
        set_error_details.extend(add_or_del)
    return set_error_details


class Transposition(object):
    '''
    Represents a transposition error that spans a sequence of potential
    boundary position.
    '''
    
    def __init__(self, start, end, boundary_type, boundaries):
        '''
        Creates an object that represents a transposition error.
        
        :param start:         Position of the beginning boundary in the \
                              transposition
        :param end:           Position of the end boundary in the transposition
        :param boundary_type: Type of boundary
        :param boundaries:    Number of boundaries involved in a transposition
        :type start:         int
        :type end:           int
        :type boundary_type: int
        :type boundaries:    int
        
        '''
        # pylint: disable=C0103
        self.start         = start
        self.end           = end
        self.boundaries    = boundaries
        self.boundary_type = boundary_type
        self.n             = (end - start) + 1
    
    def overlaps(self, other):
        '''
        Check to see if a transposition overlaps.
        '''
        overlap = False
        
        if other.boundary_type == self.boundary_type:
            if other.start >= self.start and other.start <= self.end:
                overlap = True
            elif other.end >= self.start and other.end <= self.end:
                overlap = True
            elif self.start >= other.start and self.start <= other.end:
                overlap = True
            elif self.end >= other.start and self.end <= other.end:
                overlap = True
        
        return overlap
    
    def te(self):      # pylint: disable=C0103
        '''
        Transposition error scaling function (te).
        
        .. math::
            \\text{te}(n,b)=b-(^1/_b)^{n-2} \quad \\text{where } n \geq 2 \\text{ and } b > 0
        
        :param n: PBs over which transpositions are allowed
        :param b: number of boundaries involved in the current transposition
        :type n: int
        :type b: int
        
        :returns: Scaled penalty for the transposition in question
        :rtype: :class:`decimal.Decimal`
        '''
        max_penalty = Decimal(self.boundaries*2)
        discount = (Decimal(1)/Decimal(self.boundaries*2)) ** (self.n-2)
        return max_penalty - discount
    
    def __str__(self):
        return '(%i,%i) b=%i t=%i' % (self.start, self.end, self.boundaries,
                                      self.boundary_type)
        

def set_errors_transpositions_n_edits(string_a, string_b, boundary_types, n):
    '''
    Computes edit distance of boundary strings using n-wise transpositions and
    substitutions (symmetrically).
    
    Arguments:
    :param string_a:       Boundary string
    :param string_b:       Boundary string to compare against
    :param boundary_types: Boundary types used
    :param n:              Maximum transposition window size allows (e.g. \
                                allows off by n-1)
    :type string_a:       :func:`list` of :func:`set` onjects containing \
                            :func:`int`
    :type string_b:       :func:`list` of :func:`set` onjects containing \
                            :func:`int`
    :type boundary_types: :func:`list` of :func:`int`
    :type n:                int
    
    .. seealso:: :func:`boundary_string_from_masses`
    
    :returns: edit distance, list of transpositions per position, and a list of 
        set errors (additions/substitutions)
    :rtype: func:`int`, func:`list` of class:`Transposition`, func:`list` of \
        class:`SetError`
    '''
    # pylint: disable=R0914,C0103,R0912
    
    # Identify the number of direct set errors
    set_errors = calculate_set_errors(string_a, string_b)
    
    # Define boundary type string renderer
    def render_boundary_type_string(subsequence, boundary_type):
        '''
        Creates an actual string representing boundaries of one type for a 
        subsequence of the large boundary set sequence.
        
        :param subsequence:   sequence of boundary type sets
        :param boundary_type: boundary type
        :type subsequence:    func:`list` of :func:`set` onjects containing \
                                :func:`int`
        :type boundary_type:  int
        '''
        string = ''
        for part in subsequence:
            if boundary_type in part:
                string += BOUNDARY
            else:
                string += NO_BOUNDARY
        return string
    
    set_transpositions = list()
    
    def find_overlapping_transpositions(current):
        '''
        Find transpositions which overlap each other.
        
        :param current: Transposition
        :type current:  Transposition
        
        :returns: Overlapping transpositions and the number of boundaries that \
            they span
        :rtype: :func:`list` of :class:`Transposition`, :func:`int`
        '''
        overlap = list()
        boundaries = 0
        for transposition in set_transpositions:
            if transposition.overlaps(current):
                overlap.append(transposition)
                boundaries += transposition.boundaries
        return overlap, boundaries
        
    # Identify the number of transposed set items if n >= 2
    if n >= 2:
        pairs = zip(string_a, string_b)
        test_n = range(2, n+1)
        for n_i in test_n:
            i = 0
            while i < len(pairs) - 1:
                # If out of the range
                if n_i + i > (len(pairs)):
                    break
                
                # Get subsequences
                pos_start = i
                pos_end   = (i+n_i)
                pairs_i  = pairs[pos_start:pos_end]
                subseq_a = [pair[0] for pair in pairs_i]
                subseq_b = [pair[1] for pair in pairs_i]
                
                # Look for transpositions by type
                types_transposed = list()
                for boundary_type in boundary_types:
                    string_t_a = render_boundary_type_string(subseq_a,
                                                             boundary_type)
                    string_t_b = render_boundary_type_string(subseq_b,
                                                             boundary_type)
                    boundaries = string_t_a.count(BOUNDARY)
                    
                    # Test to see if it is a potential minimal width transp
                    if string_t_b[0] != BOUNDARY and string_t_b[-1] != BOUNDARY:
                        continue
                    
                    # If can be transposed
                    if string_t_a != string_t_b and \
                        string_t_a[::-1] == string_t_b:
                        transp = Transposition(pos_start,
                                               pos_end - 1,
                                               boundary_type,
                                               boundaries)
                        types_transposed.append(transp)
                
                # If found transpositions, add to the list
                for transposition in types_transposed:
                    overlap, boundaries = find_overlapping_transpositions(
                                                                transposition)
                    
                    if len(overlap) == 0:
                        set_transpositions.append(transposition)
                    else:
                        if transposition.boundaries > boundaries:
                            # Remove overlapping
                            for transp_overlap in overlap:
                                set_transpositions.remove(transp_overlap)
                            # Add new
                            set_transpositions.append(transposition)
                    
                i += 1
    
    # Discount substitutions by transpositions
    untransposted_set_errors = list()
    transposted_set_errors   = list()
    for set_error in set_errors:
        overlap = False
        for transposition in set_transpositions:
            if set_error.overlaps(transposition):
                overlap = True
                break
        if overlap == False:
            untransposted_set_errors.append(set_error)
        else:
            transposted_set_errors.append(set_error)
    
    # Calculate distance
    distance = len(untransposted_set_errors) + len(set_transpositions)
    
    return distance, set_transpositions, untransposted_set_errors
        




'''
Tests some math related utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from decimal import Decimal
from .Math import mean, std, var, stderr


class TestMath(unittest.TestCase):
    '''
    Math utlity function tests.
    '''
    #pylint: disable=R0904,C0103
    
    def test_mean(self):
        '''
        Tests population mean.
        '''
        self.assertEqual(5, mean([2, 4, 4, 4, 5, 5, 7, 9]))
    
    def test_std(self):
        '''
        Tests population standard deviation.
        '''
        self.assertEqual(2, std([2, 4, 4, 4, 5, 5, 7, 9]))
    
    def test_var(self):
        '''
        Tests population variance.
        '''
        self.assertEqual(4, var([2, 4, 4, 4, 5, 5, 7, 9]))
    
    def test_stderr(self):
        '''
        Tests population standard error of the mean.
        '''
        self.assertEqual(Decimal('0.7071067811865475244008443622'),
                         stderr([2, 4, 4, 4, 5, 5, 7, 9]))
        



'''
Tests the WindowDiff evaluation metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from decimal import Decimal
from .Percentage import percentage, pairwise_percentage, \
    find_boundary_position_freqs
from ..data.Samples import KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, \
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT
from .. import convert_positions_to_masses



class TestPercentage(unittest.TestCase):
    '''
    Test segmentation percentage.
    '''
    # pylint: disable=R0904

    def test_identical(self):
        '''
        Test whether identical segmentations produce 1.0.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                        [1,1,1,1,1,2,2,2,3,3,3,3,3])
        segs_b = convert_positions_to_masses(
                                        [1,1,1,1,1,2,2,2,3,3,3,3,3])
        self.assertEqual(percentage(segs_a, segs_b),1.0)

    def test_no_boundaries(self):
        '''
        Test whether no segments versus some segments produce 0.0.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                        [1,1,1,1,1,1,1,1,1,1,1,1,1])
        segs_b = convert_positions_to_masses(
                                        [1,1,1,1,2,2,2,2,3,3,3,3,3])
        self.assertEqual(percentage(segs_a, segs_b),0)
        self.assertEqual(percentage(segs_b, segs_a),0)

    def test_all_boundaries(self):
        '''
        Test whether all segments versus some segments produces 2/12, or 0.167.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                    [1,2,3,4,5,6,7,8,9,10,11,12,13])
        segs_b = convert_positions_to_masses(
                                    [1,1,1,1,2,2,2,2,3,3,3,3,3])
        self.assertEqual(percentage(segs_a, segs_b),
                         Decimal('0.1666666666666666666666666667'))
        self.assertEqual(percentage(segs_b, segs_a),
                         Decimal('0.1666666666666666666666666667'))

    def test_all_and_no_boundaries(self):
        '''
        Test whether all segments versus no segments produces 0.0.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                    [1,2,3,4,5,6,7,8,9,10,11,12,13])
        segs_b = convert_positions_to_masses(
                                    [1,1,1,1,1,1,1,1,1,1,1,1,1])
        self.assertEqual(percentage(segs_a, segs_b),0)
        self.assertEqual(percentage(segs_b, segs_a),0)

    def test_translated_boundary(self):
        '''
        Test whether 2/3 total segments participate in  mis-alignment produces
        0.33.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                        [1,1,1,1,1,2,2,2,3,3,3,3,3])
        segs_b = convert_positions_to_masses(
                                        [1,1,1,1,2,2,2,2,3,3,3,3,3])
        self.assertEqual(percentage(segs_a, segs_b),
                         Decimal('0.3333333333333333333333333333'))
        self.assertEqual(percentage(segs_b, segs_a),
                         Decimal('0.3333333333333333333333333333'))
    
    def test_extra_boundary(self):
        '''
        Test whether 1/3 segments that are non-existent produces 0.66.
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                        [1,1,1,1,1,2,2,2,3,3,3,3,3])
        segs_b = convert_positions_to_masses(
                                        [1,1,1,1,1,2,3,3,4,4,4,4,4])
        self.assertEqual(percentage(segs_a, segs_b),
                         Decimal('0.6666666666666666666666666667'))
        self.assertEqual(percentage(segs_b, segs_a),
                         Decimal('0.6666666666666666666666666667'))
    
    def test_full_miss_and_misaligned(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.25. 
        '''
        # pylint: disable=C0324
        segs_a = convert_positions_to_masses(
                                        [1,1,1,1,2,2,2,2,3,3,3,3,3])
        segs_b = convert_positions_to_masses(
                                        [1,1,1,1,1,2,3,3,4,4,4,4,4])
        self.assertEqual(percentage(segs_a, segs_b), Decimal('0.25'))
        self.assertEqual(percentage(segs_b, segs_a), Decimal('0.25'))


class TestPairwisePercentage(unittest.TestCase):
    # pylint: disable=R0904
    '''
    Test permuted pairwise percentage.
    '''
    
    def test_kazantseva2012_g5(self):
        '''
        Calculate permuted pairwise percentage on Group 5 from the dataset
        collected in Kazantseva (2012).
        '''
        self.assertEqual(pairwise_percentage(KAZANTSEVA2012_G5),
                         (Decimal('0.1621263635243898401793138635'),
                          Decimal('0.1788409781886208812486660585'),
                          Decimal('0.03198409547946276978304443503'),
                          Decimal('0.03650576180519474391025947712')))
    
    def test_kazantseva2012_g2(self):
        '''
        Calculate mean permuted pairwise percentage on Group 2 from the dataset
        collected in Kazantseva (2012).
        '''
        self.assertEqual(pairwise_percentage(KAZANTSEVA2012_G2),
                         (Decimal('0.3398087832646656176067940768'),
                          Decimal('0.1948481072924021072633034332'),
                          Decimal('0.03796578491543144325163024138'),
                          Decimal('0.02515478248611697670879150623')))
    
    def test_large_disagreement(self):
        '''
        Calculate mean permuted pairwise percentage on a theoretical dataset
        containing large disagreement.
        '''
        self.assertEqual(pairwise_percentage(LARGE_DISAGREEMENT),
                         (0.0,
                          0.0,
                          0.0,
                          0.0))
    
    def test_complete_agreement(self):
        '''
        Calculate mean permuted pairwise percentage on a theoretical dataset
        containing complete agreement.
        '''
        self.assertEqual(pairwise_percentage(COMPLETE_AGREEMENT),
                         (1.0,
                          0.0,
                          0.0,
                          0.0))


class TestPercentageUtils(unittest.TestCase):
    # pylint: disable=R0904
    '''
    Test utility functions used to calculate percentage.
    '''
    
    def test_find_seg_positions(self):
        '''
        Test segmentation position frequency counting.
        '''
        # pylint: disable=C0324
        seg_positions = find_boundary_position_freqs([[1,2,3,3,2,1],
                                                      [1,2,2,4,2,1]])
        self.assertEqual(seg_positions, { 1: 2,
                                          3: 2,
                                          5: 1,
                                          6: 1,
                                          9: 2,
                                         11: 2})




'''
Math utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
from decimal import Decimal


def mean(values):
    '''
    Calculates the mean of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Mean.
    :rtype: :class:`decimal.Decimal`
    '''
    summation = Decimal(0)
    for value in values:
        summation += value
    return summation / len(values)
    
    
def var(values):
    '''
    Calculates the population variance of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Variance.
    :rtype: :class:`decimal.Decimal`
    '''
    mean_value = mean(values)
    summation = Decimal(0)
    for value in values:
        summation += (value - mean_value) ** 2
    return summation / len(values)


def std(values):
    '''
    Calculates the population standard deviation of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Standard deviation.
    :rtype: :class:`decimal.Decimal`
    '''
    return var(values).sqrt()


def stderr(values):
    '''
    Calculates the population standard error of the mean of a list of numeric
    values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Standard error of the mean.
    :rtype: :class:`decimal.Decimal`
    '''
    return std(values) / Decimal(len(values)).sqrt()




'''
Data I/O package.  Used to import and export data to and from TSV and JSON
files.

.. seealso:: File format documentation in: `Segmentation Representation \
Specification <http://nlp.chrisfournier.ca/publications/#seg_spec>`_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import os
import csv
import json
from .. import convert_positions_to_masses

RESULTS             = ['summary', 'tsv']

DEFAULT_DELIMITER   = '\t'


def load_tests(loader, tests, pattern):
    '''
    A ``load_tests()`` function utilizing the default loader
    :func:`segeval.Utils.default_load_tests`.
    
    .. seealso:: The `load_tests protocol <http://docs.python.org/library/\
    unittest.html#load-tests-protocol>`_.
    '''
    #pylint: disable=W0613
    from ..Utils import default_load_tests
    return default_load_tests(__file__, loader, tests)


class DataIOError(Exception):
    '''
    Indicates that an input processing error has occurred.
    '''
    
    def __init__(self, message, exception):
        '''
        Initializer.
        
        :param message: Explanation for the exception.
        :type message: str
        '''
        Exception.__init__(self, message, exception)
    

def input_linear_mass_tsv(tsv_filename, delimiter=DEFAULT_DELIMITER):
    '''
    Load a linear segmentation mass TSV file.
    
    :param tsv_filename: path to the mass file containing segment mass codings.
    :param delimiter:    the delimiter used when reading a TSV file (by default,
                         a tab, but it can also be a comma, whitespace, etc.
    :type tsv_filename: str
    :type delimiter: str
    
    :returns: Segmentation mass codings.
    :rtype: :func:`dict`
    '''
    # List version of file
    header = []
    segment_masses = dict()
    # Open file
    csv_file = open(tsv_filename, 'rU')
    # Read in file
    try:
        reader = csv.reader(csv_file, delimiter=delimiter)
        for i, row in enumerate(reader):
            # Read annotators from header
            if i == 0:
                for item in row[1:]:
                    header.append(item)
            # Read data
            else:
                coder = None
                for j, col in enumerate(row):
                    # Skip the first col
                    if j == 0:
                        coder = str(col)
                        segment_masses[coder] = list()
                    elif j > 0:
                        segment_masses[coder].append(int(col))
    # pylint: disable=C0103
    except Exception as exception:
        raise DataIOError('Error occurred processing file: %s' \
                                      % tsv_filename, exception)
    finally:
        csv_file.close()
    return segment_masses


def input_linear_positions_tsv(tsv_filename, delimiter=DEFAULT_DELIMITER):
    '''
    Load a segment position TSV file.
    
    :param csv_filename: path to the mass file containing segment position
                         codings.
    :param delimiter:    the delimiter used when reading a TSV file (by default,
                         a tab, but it can also be a comma, whitespace, etc.
    :type csv_filename: str
    :type delimiter: str
    
    .. deprecated:: 1.0
    
    .. warning:: This i/o function is for legacy files only and will be removed
        in later versions.
    
    :returns: Segmentation mass codings.
    :rtype: :func:`dict`
    '''
    coder_positions = input_linear_mass_tsv(tsv_filename, delimiter)
    # Convert each segment position to masses
    for coder, positions in coder_positions.items():
        coder_positions[coder] = convert_positions_to_masses(positions)
    # Return
    return coder_positions


def input_linear_mass_json(json_filename):
    '''
    Load a segment mass JSON file.
    
    :param json_filename: path to the mass file containing segment position
                          codings.
    :type json_filename: str
    
    :returns: Segmentation mass codings.
    :rtype: :func:`dict`
    
    .. seealso:: `JSON (JavaScript Object Notation) <http://www.json.org/>`_.
    '''
    codings = dict()
    data = dict()
    # Open file
    json_file = open(json_filename, 'rU')
    # Read in file
    try:
        data = json.load(json_file)
    except Exception as exception:
        raise DataIOError('Error occurred processing file: %s' \
                                      % json_filename, exception)
    # Check type
    if 'segmentation_type' in data:
        if data['segmentation_type'] != 'linear':
            raise DataIOError(
                'Segmentation type \'linear\' expected, but encountered %s' % \
                data['segmentation_type'])
    # Remove the metadata layer
    if 'codings' in data:
        data = data['codings']
    else:
        data = data
    # Convert coder labels into strings
    for key, value in data.items():
        codings[key] = value
    # Return
    return codings


FILETYPE_TSV  = 'tsv'
FILETYPE_JSON = 'json'

EXT = 'ext'
FNC = 'fnc'
FILETYPES           = {FILETYPE_TSV  : {EXT : ['.tsv', '.csv'],
                                        FNC : input_linear_mass_tsv},
                       FILETYPE_JSON : {EXT : ['.json', '.jsn'],
                                        FNC : input_linear_mass_json}}
FILETYPES_DEFAULT   = FILETYPE_JSON


def load_nested_folders_dict(containing_dir, filetype):
    '''
    Loads TSV files from a file directory structure, which reflects the
    directory structure in nested :func:`dict` with each directory name
    representing a key in these :func:`dict`.
    
    :param containing_dir: Root directory containing sub-directories which 
                           contain segmentation files.
    :param filetype:       File type to load (e.g., json or tsv).
    :type containing_dir: str
    :type filetype: str
    
    :returns: Segmentation mass codings.
    :rtype: :func:`dict`
    '''
    allowable_extensions = list(FILETYPES[filetype][EXT])
    fnc_load = FILETYPES[filetype][FNC]
    data = dict()
    datafile_found = False
    # List of entries
    files = dict()
    dirs = dict()
    # For each filesystem item
    for name in os.listdir(containing_dir):
        path = os.path.join(containing_dir, name)
        # Found a directory
        if os.path.isdir(path):
            dirs[name] = path
        # Found a file
        elif os.path.isfile(path):
            name, ext = os.path.splitext(name)
            if len(ext) > 0 and ext.lower() in allowable_extensions:
                files[name] = path
                datafile_found = True
    # If a data file was found
    if datafile_found:
        # If TSV files were found, load
        for name, filepath in files.items():
            data[name] = fnc_load(filepath)
    else:
        # If only dirs were found, recurse
        for name, dirpath in dirs.items():
            data[name] = load_nested_folders_dict(dirpath, filetype)
    return data


def load_file(args):
    '''
    Load a file or set of directories from command line arguments.
    
    :param args: Command line arguments
    :type args: dict
    
    :returns: The loaded values and whether a file was loaded or not.
    :rtype: :func:`dict`, :func:`bool`
    '''
    values = None
    
    input_path = args['input'][0]
    is_file = os.path.isfile(input_path)
    
    filetype = args['format']
    
    # Load file or dir
    if is_file:
        values = FILETYPES[filetype][FNC](input_path)
        values = {'item' : values}
    else:
        values = load_nested_folders_dict(input_path, filetype)
    
    return values, is_file


def parser_add_file_support(parser):
    '''
    Add support for file input and output parameters to an argument parser.
    
    :param parser: Argument parser
    :type parser: argparse.ArgumentParser
    '''
    
    parser.add_argument('-o', '--output',
                        type=str,
                        nargs=1,
                        required=False,
                        help='Output file or directory. If not specified, a '+\
                        'summary or results is printed to the console.')
    
    parser.add_argument('input',
                        type=str,
                        nargs=1,
                        action='store',
                        help='Input file or directory')
    
    parser.add_argument('-f', '--format',
                        type=str,
                        default=FILETYPES_DEFAULT,
                        choices=FILETYPES.keys(),
                        help='Input file format; default is %s' % \
                            FILETYPES_DEFAULT)
    
    parser.add_argument('-d', '--delimiter',
                        type=str,
                        default='\t',
                        help='Delimiting character for input TSV files; '+\
                        'ignored if JSON is specified, default is a tab '+\
                        'character')




'''
Tests the segmentation versions of Cohen's and Fleiss' Kappa.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================
import unittest
from decimal import Decimal
from .Kappa import cohen_kappa, fleiss_kappa, mean_fleiss_kappa
from ..data.Samples import KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, \
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT


class TestKappa(unittest.TestCase):
    '''
    Test segmentation versions of Cohen's Kappa and Fleiss' Multi-Kappa.
    '''
    # pylint: disable=R0904
    
    SKIP = False

    def test_fleiss_kappa_g5(self):
        '''
        Test Kappa upon Group 5 of Kazantseva (2012) data.
        '''
        if TestKappa.SKIP:
            return
        data = KAZANTSEVA2012_G5
        self.assertEqual(fleiss_kappa(data),
                         Decimal('0.8198449828922561273654074436'))


    def test_fleiss_kappa_g5_ch1(self):
        '''
        Test Kappa upon Group 5, Chapter 1, of Kazantseva (2012) data.
        '''
        if TestKappa.SKIP:
            return
        data = KAZANTSEVA2012_G5['ch1']
        data = {'ch1' : data}
        self.assertEqual(fleiss_kappa(data),
                         Decimal('0.7462686567164179104477611940'))


    def test_fleiss_kappa_g2(self):
        '''
        Test Kappa upon Group 2 of Kazantseva (2012) data.
        '''
        if TestKappa.SKIP:
            return
        data = KAZANTSEVA2012_G2
        self.assertEqual(fleiss_kappa(data),
                         Decimal('0.8865951832180913251160770952'))


    def test_fleiss_kappa_g2_ch2(self):
        '''
        Test Kappa upon Group 2, Chapter 2, of Kazantseva (2012) data.
        '''
        if TestKappa.SKIP:
            return
        data = KAZANTSEVA2012_G2['ch2']
        data = {'ch2' : data}
        self.assertEqual(fleiss_kappa(data),
                         Decimal('0.8840057636887608069164265130'))
        

    def test_fleiss_kappa_disagree(self):
        '''
        Test Kappa upon a hypothetical dataset containing large disagreement.
        '''
        if TestKappa.SKIP:
            return
        data = LARGE_DISAGREEMENT
        self.assertEqual(fleiss_kappa(data),
                         Decimal('-0.05952156715012243360331512524'))
    
    
    def test_cohen_kappa(self):
        '''
        Test Cohen's and Fleiss' Kappa.
        '''
        if TestKappa.SKIP:
            return
        data1 = {'i1' : {'c1' : [2, 8, 2, 1],
                         'c2' : [2, 1, 7, 2, 1]}}
        kappa1  = cohen_kappa(data1)
        kappa1f = fleiss_kappa(data1)
        self.assertEqual(kappa1,
                         Decimal('0.9032258064516129032258064517'))
        self.assertEqual(kappa1, kappa1f)
        data2 = {'i1' : {'c1' : [2, 8, 2, 1],
                         'c2' : [11, 2]}}
        kappa2  = cohen_kappa(data2)
        kappa2f = fleiss_kappa(data2)
        self.assertEqual(kappa2,
                         Decimal('0.7352941176470588235294117647'))
        self.assertEqual(kappa2, kappa2f)
        self.assertTrue(kappa2 < kappa1)
    
    
    def test_cohen_kappa_complete(self):
        '''
        Test Kappa upon a hypothetical dataset containing complete agreement.
        '''
        if TestKappa.SKIP:
            return
        # pylint: disable=C0324
        data_complete = {'i1' : {'c1' : [2,8,2,1],
                                 'c2' : [2,8,2,1]}}
        kappa = cohen_kappa(data_complete)
        self.assertEqual(kappa, 1.0)
    
    
    def test_fleiss_kappa_complete(self):
        '''
        Test Kappa upon a hypothetical dataset containing complete agreement.
        '''
        if TestKappa.SKIP:
            return
        data_complete = COMPLETE_AGREEMENT
        kappa = fleiss_kappa(data_complete)
        self.assertEqual(kappa, 1.0)


    def test_mean_fleiss_kappa(self):
        '''
        Calculate mean Fleiss' Kappa on Group 2 and 5 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertEqual(mean_fleiss_kappa(
                                {'g2' : KAZANTSEVA2012_G2,
                                 'g5' : KAZANTSEVA2012_G5}),
                         (Decimal('0.8532200830551737262407422694'),
                          Decimal('0.03337510016291759887533482579'),
                          Decimal('0.001113897310884782334990518712'),
                          Decimal('0.01179987982398964061050486219')))




'''
Segmentation Similarity (S) package [FournierInkpen2012]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
#===============================================================================
# Copyright (c) 2011-2012, Chris Fournier
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#       
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#===============================================================================


def load_tests(loader, tests, pattern):
    '''
    A ``load_tests()`` function utilizing the default loader
    :func:`segeval.Utils.default_load_tests`.
    
    .. seealso:: The `load_tests protocol <http://docs.python.org/library/\
    unittest.html#load-tests-protocol>`_.
    '''
    #pylint: disable=W0613
    from ..Utils import default_load_tests
    return default_load_tests(__file__, loader, tests)

